from typing import Optional, List
import logging
from collections import defaultdict

import pyvex
import ailment

from ...block import Block
from ...errors import AngrVariableRecoveryError, SimEngineError
from ...knowledge_plugins import Function
from ...sim_variable import SimStackVariable, SimRegisterVariable, SimVariable
from ...engines.vex.claripy.irop import vexop_to_simop
from ..forward_analysis import ForwardAnalysis, FunctionGraphVisitor
from ..typehoon.typevars import Equivalence, TypeVariable
from .variable_recovery_base import VariableRecoveryBase, VariableRecoveryStateBase
from .engine_vex import SimEngineVRVEX
from .engine_ail import SimEngineVRAIL

l = logging.getLogger(name=__name__)


class ProcessorState:

    __slots__ = ['_arch', 'sp_adjusted', 'sp_adjustment', 'bp_as_base', 'bp']

    def __init__(self, arch):
        self._arch = arch
        # whether we have met the initial stack pointer adjustment
        self.sp_adjusted = None
        # how many bytes are subtracted from the stack pointer
        self.sp_adjustment = arch.bytes if arch.call_pushes_ret else 0
        # whether the base pointer is used as the stack base of the stack frame or not
        self.bp_as_base = None
        # content of the base pointer
        self.bp = None

    def copy(self):
        s = ProcessorState(self._arch)
        s.sp_adjusted = self.sp_adjusted
        s.sp_adjustment = self.sp_adjustment
        s.bp_as_base = self.bp_as_base
        s.bp = self.bp
        return s

    def merge(self, other):
        if not self == other:
            l.warning("Inconsistent merge: %s %s ", self, other)

        # FIXME: none of the following logic makes any sense...
        if other.sp_adjusted is True:
            self.sp_adjusted = True
        self.sp_adjustment = max(self.sp_adjustment, other.sp_adjustment)
        if other.bp_as_base is True:
            self.bp_as_base = True
        if self.bp is None:
            self.bp = other.bp
        elif other.bp is not None:  # and self.bp is not None
            if self.bp == other.bp:
                pass
            else:
                if type(self.bp) is int and type(other.bp) is int:
                    self.bp = max(self.bp, other.bp)
                else:
                    self.bp = None
        return self

    def __eq__(self, other):
        if not isinstance(other, ProcessorState):
            return False
        return (self.sp_adjusted == other.sp_adjusted and
                self.sp_adjustment == other.sp_adjustment and
                self.bp == other.bp and
                self.bp_as_base == other.bp_as_base)

    def __repr__(self):
        return "<ProcessorState %s%#x%s %s>" % (self.bp, self.sp_adjustment,
            " adjusted" if self.sp_adjusted else "", self.bp_as_base)


class VariableRecoveryFastState(VariableRecoveryStateBase):
    """
    The abstract state of variable recovery analysis.

    :ivar KeyedRegion stack_region: The stack store.
    :ivar KeyedRegion register_region:  The register store.
    """
    def __init__(self, block_addr, analysis, arch, func, stack_region=None, register_region=None, global_region=None,
                 typevars=None, type_constraints=None, delayed_type_constraints=None, processor_state=None):

        super().__init__(block_addr, analysis, arch, func, stack_region=stack_region, register_region=register_region,
                         global_region=global_region, typevars=typevars, type_constraints=type_constraints,
                         delayed_type_constraints=delayed_type_constraints)

        self.processor_state = ProcessorState(self.arch) if processor_state is None else processor_state

    def __repr__(self):
        return "<VRAbstractState@%#x: %d register variables, %d stack variables>" % (
            self.block_addr, len(self.register_region), len(self.stack_region))

    def __eq__(self, other):
        if type(other) is not VariableRecoveryFastState:
            return False
        return self.stack_region == other.stack_region and self.register_region == other.register_region

    def copy(self):

        state = VariableRecoveryFastState(
            self.block_addr,
            self._analysis,
            self.arch,
            self.function,
            stack_region=self.stack_region.copy(),
            register_region=self.register_region.copy(),
            global_region=self.global_region.copy(),
            typevars=self.typevars.copy(),
            type_constraints=self.type_constraints.copy(),
            processor_state=self.processor_state.copy(),
            delayed_type_constraints=self.delayed_type_constraints.copy(),
        )

        return state

    def merge(self, other, successor=None):
        """
        Merge two abstract states.

        For any node A whose dominance frontier that the current node (at the current program location) belongs to, we
        create a phi variable V' for each variable V that is defined in A, and then replace all existence of V with V'
        in the merged abstract state.

        :param VariableRecoveryState other: The other abstract state to merge.
        :return:                            The merged abstract state.
        :rtype:                             VariableRecoveryState
        """

        replacements = {}
        if successor in self.dominance_frontiers:
            replacements = self._make_phi_variables(successor, self, other)

        merged_stack_region = self.stack_region.copy().replace(replacements).merge(other.stack_region,
                                                                                   replacements=replacements)
        merged_register_region = self.register_region.copy().replace(replacements).merge(other.register_region,
                                                                                         replacements=replacements)
        merged_global_region = self.global_region.copy().merge(other.global_region)
        merged_typevars = self.typevars.merge(other.typevars)
        merged_typeconstraints = self.type_constraints.copy() | other.type_constraints
        delayed_typeconstraints = self.delayed_type_constraints.copy()
        for v, cons in other.delayed_type_constraints.items():
            delayed_typeconstraints[v] |= cons
        # add subtype constraints for all replacements
        for v0, v1 in replacements.items():
            # v0 will be replaced by v1
            if not merged_typevars.has_type_variable_for(v1, None):
                merged_typevars.add_type_variable(v1, None, TypeVariable())
            if not merged_typevars.has_type_variable_for(v0, None):
                merged_typevars.add_type_variable(v0, None, TypeVariable())
            # Assuming v2 = phi(v0, v1), then we know that v0_typevar == v1_typevar == v2_typevar
            # However, it's possible that neither v0 nor v1 will ever be used in future blocks, which not only makes
            # this phi function useless, but also leads to the incorrect assumption that v1_typevar == v2_typevar.
            # Hence, we delay the addition of the equivalence relationship into the type constraints. It is only added
            # when v1 (the new variable that will end up in the state) is ever used in the future.

            # create an equivalence relationship
            equivalence = Equivalence(merged_typevars.get_type_variable(v1, None),
                                      merged_typevars.get_type_variable(v0, None)
                                      )
            delayed_typeconstraints[v1].add(equivalence)

        state = VariableRecoveryFastState(
            successor,
            self._analysis,
            self.arch,
            self.function,
            stack_region=merged_stack_region,
            register_region=merged_register_region,
            global_region=merged_global_region,
            typevars=merged_typevars,
            type_constraints=merged_typeconstraints,
            delayed_type_constraints=delayed_typeconstraints,
            processor_state=self.processor_state.copy().merge(other.processor_state),
        )

        return state

    #
    # Util methods
    #

    def _normalize_register_offset(self, offset):  #pylint:disable=no-self-use

        # TODO:

        return offset

    def _to_signed(self, n):

        if n >= 2 ** (self.arch.bits - 1):
            # convert it to a negative number
            return n - 2 ** self.arch.bits

        return n


class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  #pylint:disable=abstract-method
    """
    Recover "variables" from a function by keeping track of stack pointer offsets and pattern matching VEX statements.

    If calling conventions are recovered prior to running VariableRecoveryFast, variables can be recognized more
    accurately. However, it is not a requirement.
    """

    def __init__(self, func, func_graph=None, max_iterations=1, low_priority=False, track_sp=True,
                 func_args: Optional[List[SimVariable]]=None):
        """

        :param knowledge.Function func:  The function to analyze.
        :param int max_iterations:
        :param clinic:
        """

        function_graph_visitor = FunctionGraphVisitor(func, graph=func_graph)

        # Make sure the function is not empty
        if not func.block_addrs_set or func.startpoint is None:
            raise AngrVariableRecoveryError("Function %s is empty." % repr(func))

        VariableRecoveryBase.__init__(self, func, max_iterations)
        ForwardAnalysis.__init__(self, order_jobs=True, allow_merging=True, allow_widening=False,
                                 graph_visitor=function_graph_visitor)

        self._low_priority = low_priority
        self._job_ctr = 0
        self._track_sp = track_sp
        self._func_args = func_args

        self._ail_engine = SimEngineVRAIL(self.project, self.kb)
        self._vex_engine = SimEngineVRVEX(self.project, self.kb)

        self._node_iterations = defaultdict(int)

        self._node_to_cc = { }
        self.var_to_typevar = { }
        self.type_constraints = None

        self._analyze()

        # cleanup (for cpython pickle)
        self._ail_engine = None
        self._vex_engine = None

    #
    # Main analysis routines
    #

    def _pre_analysis(self):

        self.type_constraints = set()

        self.initialize_dominance_frontiers()

        if self._track_sp:
            # initialize node_to_cc map
            function_nodes = [n for n in self.function.transition_graph.nodes() if isinstance(n, Function)]
            # all nodes that end with a call must be in the _node_to_cc dict
            for func_node in function_nodes:
                for callsite_node in self.function.transition_graph.predecessors(func_node):
                    if func_node.calling_convention is None:
                        l.warning("Unknown calling convention for %r.", func_node)
                        self._node_to_cc[callsite_node.addr] = None
                    else:
                        self._node_to_cc[callsite_node.addr] = func_node.calling_convention

    def _pre_job_handling(self, job):
        self._job_ctr += 1
        if self._low_priority:
            self._release_gil(self._job_ctr, 5, 0.000001)

    def _initial_abstract_state(self, node):

        # annotate the stack pointer
        # concrete_state.regs.sp = concrete_state.regs.sp.annotate(StackLocationAnnotation(8))

        # give it enough stack space
        # concrete_state.regs.bp = concrete_state.regs.sp + 0x100000

        state = VariableRecoveryFastState(node.addr, self, self.project.arch, self.function,
                                          )
        # put a return address on the stack if necessary
        if self.project.arch.call_pushes_ret:
            ret_addr_offset = self.project.arch.bytes
            ret_addr_var = SimStackVariable(ret_addr_offset, self.project.arch.bytes, base='bp', name='ret_addr',
                                            region=self.function.addr, category='return_address',
                                            )
            state.stack_region.add_variable(ret_addr_offset, ret_addr_var)

        if self._func_args:
            for arg in self._func_args:
                if isinstance(arg, SimRegisterVariable):
                    state.register_region.set_variable(arg.reg, arg)
                    self.variable_manager[self.function.addr].add_variable('register', arg.reg, arg)
                elif isinstance(arg, SimStackVariable):
                    state.stack_region.set_variable(arg.offset, arg)
                    self.variable_manager[self.function.addr].add_variable('stack', arg.offset, arg)
                else:
                    raise TypeError("Unsupported function argument type %s." % type(arg))

        return state

    def _merge_states(self, node, *states):

        return states[0].merge(states[1], successor=node.addr)

    def _run_on_node(self, node, state):
        """


        :param angr.Block node:
        :param VariableRecoveryState state:
        :return:
        """

        input_state = state  # make it more meaningful

        if type(node) is ailment.Block:
            # AIL mode
            block = node
        else:
            # VEX mode, get the block again
            block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)

        if node.addr in self._instates:
            prev_state = self._instates[node.addr]
            if input_state == prev_state:
                l.debug('Skip node %#x as we have reached a fixed-point', node.addr)
                return False, input_state
            else:
                l.debug('Merging input state of node %#x with the previous state.', node.addr)
                input_state = prev_state.merge(input_state, successor=node.addr)

        state = input_state.copy()
        state.block_addr = node.addr
        self._instates[node.addr] = input_state

        if self._node_iterations[node.addr] >= self._max_iterations:
            l.debug('Skip node %#x as we have iterated %d times on it.', node.addr, self._node_iterations[node.addr])
            return False, state

        self._process_block(state, block)

        self._outstates[node.addr] = state

        self._node_iterations[node.addr] += 1
        self.type_constraints |= state.type_constraints
        self.var_to_typevar.update(state.typevars._typevars)

        return True, state

    def _intra_analysis(self):
        pass

    def _post_analysis(self):
        self.variable_manager['global'].assign_variable_names(labels=self.kb.labels)
        self.variable_manager[self.function.addr].assign_variable_names()

        for addr, state in self._outstates.items():
            self.variable_manager[self.function.addr].set_live_variables(addr,
                                                                         state.register_region,
                                                                         state.stack_region
                                                                         )

    #
    # Private methods
    #

    @staticmethod
    def _get_irconst(value, size):
        mapping = {
            1: pyvex.const.U1,
            8: pyvex.const.U8,
            16: pyvex.const.U16,
            32: pyvex.const.U32,
            64: pyvex.const.U64,
            128: pyvex.const.V128,
            256: pyvex.const.V256,
        }
        if size not in mapping:
            raise TypeError("Unsupported size %d." % size)
        return mapping.get(size)(value)

    def _peephole_optimize(self, block: Block):

        # find regN = xor(regN, regN) and replace it with PUT(regN) = 0
        i = 0
        while i < len(block.vex.statements) - 3:
            stmt0 = block.vex.statements[i]
            next_i = i + 1
            if isinstance(stmt0, pyvex.IRStmt.WrTmp) and isinstance(stmt0.data, pyvex.IRStmt.Get):
                stmt1 = block.vex.statements[i + 1]
                if isinstance(stmt1, pyvex.IRStmt.WrTmp) and isinstance(stmt1.data, pyvex.IRStmt.Get):
                    next_i = i + 2
                    if stmt0.data.offset == stmt1.data.offset and stmt0.data.ty == stmt1.data.ty:
                        next_i = i + 3
                        reg_offset = stmt0.data.offset
                        tmp0 = stmt0.tmp
                        tmp1 = stmt1.tmp
                        stmt2 = block.vex.statements[i + 2]
                        if isinstance(stmt2, pyvex.IRStmt.WrTmp) and isinstance(stmt2.data, pyvex.IRExpr.Binop):
                            if isinstance(stmt2.data.args[0], pyvex.IRExpr.RdTmp) \
                                    and isinstance(stmt2.data.args[1], pyvex.IRExpr.RdTmp) \
                                    and {stmt2.data.args[0].tmp, stmt2.data.args[1].tmp} == {tmp0, tmp1}\
                                    and vexop_to_simop(stmt2.data.op)._generic_name == 'Xor':
                                # found it!
                                # make a copy so we don't trash the cached VEX IRSB
                                block._vex = block.vex.copy()
                                block.vex.statements[i] = pyvex.IRStmt.NoOp()
                                block.vex.statements[i + 1] = pyvex.IRStmt.NoOp()
                                zero = pyvex.IRExpr.Const(self._get_irconst(0, block.vex.tyenv.sizeof(tmp0)))
                                block.vex.statements[i + 2] = pyvex.IRStmt.Put(zero, reg_offset)
            i = next_i
        return block

    def _process_block(self, state, block):  # pylint:disable=no-self-use
        """
        Scan through all statements and perform the following tasks:
        - Find stack pointers and the VEX temporary variable storing stack pointers
        - Selectively calculate VEX statements
        - Track memory loading and mark stack and global variables accordingly

        :param angr.Block block:
        :return:
        """

        l.debug('Processing block %#x.', block.addr)

        if isinstance(block, Block):
            try:
                _ = block.vex
            except SimEngineError:
                # the block does not exist or lifting failed
                return
            block = self._peephole_optimize(block)

        processor = self._ail_engine if isinstance(block, ailment.Block) else self._vex_engine
        processor.process(state, block=block, fail_fast=self._fail_fast)

        if self._track_sp and block.addr in self._node_to_cc:
            # readjusting sp at the end for blocks that end in a call
            adjusted = False
            if state.processor_state.sp_adjustment is None:
                state.processor_state.sp_adjustment = 0
            cc = self._node_to_cc[block.addr]
            if cc is not None and cc.sp_delta is not None:
                state.processor_state.sp_adjustment += cc.sp_delta
                state.processor_state.sp_adjusted = True
                adjusted = True
                l.debug('Adjusting stack pointer at end of block %#x with offset %+#x.',
                        block.addr, state.processor_state.sp_adjustment)

            if not adjusted:
                # make a guess
                # of course, this will fail miserably if the function called is not cdecl
                if self.project.arch.call_pushes_ret:
                    state.processor_state.sp_adjustment += self.project.arch.bytes
                    state.processor_state.sp_adjusted = True


from angr.analyses import AnalysesHub
AnalysesHub.register_default('VariableRecoveryFast', VariableRecoveryFast)



import logging
from typing import Union

import pypcode
from pypcode import OpCode, VarnodeData, PcodeOpRaw
import claripy
from claripy.ast.bv import BV

from ..engine import SimEngineBase
from ...utils.constants import DEFAULT_STATEMENT
from .lifter import IRSB, PcodeInstruction
from .behavior import OpBehavior

l = logging.getLogger(__name__)

class PcodeEmulatorMixin(SimEngineBase):
    """
    Mixin for P-Code execution.
    """

    _current_ins:      Union[PcodeInstruction, None]
    _current_op:       Union[PcodeOpRaw, None]
    _current_behavior: Union[OpBehavior, None]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_ins = None
        self._current_op = None
        self._current_behavior = None
        self._special_op_handlers = {
            OpCode.CPUI_LOAD:       self._execute_load,
            OpCode.CPUI_STORE:      self._execute_store,
            OpCode.CPUI_BRANCH:     self._execute_branch,
            OpCode.CPUI_CBRANCH:    self._execute_cbranch,
            OpCode.CPUI_BRANCHIND:  self._execute_branchind,
            OpCode.CPUI_CALL:       self._execute_call,
            OpCode.CPUI_CALLIND:    self._execute_callind,
            OpCode.CPUI_CALLOTHER:  self._execute_callother,
            OpCode.CPUI_RETURN:     self._execute_ret,
            OpCode.CPUI_MULTIEQUAL: self._execute_multiequal,
            OpCode.CPUI_INDIRECT:   self._execute_indirect,
            OpCode.CPUI_SEGMENTOP:  self._execute_segment_op,
            OpCode.CPUI_CPOOLREF:   self._execute_cpool_ref,
            OpCode.CPUI_NEW:        self._execute_new,
        }

    def handle_pcode_block(self, irsb: IRSB) -> None:
        """
        Execute a single IRSB.

        :param irsb: The block to be executed.
        """
        self.irsb = irsb
        # Hack on a handler here to track whether exit has been handled or not
        # FIXME: Vex models this as a known exit statement, which we should also
        # do here. For now, handle it this way.
        self.state.scratch.exit_handled = False

        fallthru_addr = None
        for i, ins in enumerate(irsb._instructions):
            l.debug("Executing machine instruction @ %#x (%d of %d)", ins.addr.getOffset(), i+1, len(irsb._instructions))

            # Execute a single instruction of the emulated machine
            self._current_ins = ins
            self.state.scratch.ins_addr = self._current_ins.addr.getOffset()

            # FIXME: Hacking this on here but ideally should use "scratch".
            self._pcode_tmps = {}  # FIXME: Consider alignment requirements

            for op in self._current_ins.ops:
                self._current_op = op
                self._current_behavior = irsb.behaviors.get_behavior_for_opcode(self._current_op.getOpcode())
                l.debug("Executing p-code op: %s", self._current_ins.pp_op_str(self._current_op))
                self._execute_current_op()

            self._current_op = None
            self._current_behavior = None
            fallthru_addr = ins.addr.getOffset() + ins.length

        if not self.state.scratch.exit_handled:
            self.successors.add_successor(
                self.state,
                fallthru_addr,
                self.state.scratch.guard,
                "Ijk_Boring",
                exit_stmt_idx=DEFAULT_STATEMENT,
                exit_ins_addr=self.state.scratch.ins_addr,
            )

    def _execute_current_op(self) -> None:
        """
        Execute the current p-code operation.
        """
        assert self._current_behavior is not None

        if self._current_behavior.is_special:
            self._special_op_handlers[self._current_behavior.opcode]()
        elif self._current_behavior.is_unary:
            self._execute_unary()
        else:
            self._execute_binary()

    def _map_register_name(self, var_data: VarnodeData) -> int:
        """
        Map SLEIGH register offset to ArchInfo register offset based on name.

        :param var_data: The varnode to translate.
        :return:         The register file offset.
        """
        # FIXME: Will need performance optimization
        # FIXME: Should not get trans object this way. Moreover, should have a faster mapping method than going through trans
        trans = self.irsb._instructions[0].trans
        reg_name = trans.getRegisterName(var_data.space, var_data.offset, var_data.size)
        try:
            reg_offset = self.state.project.arch.get_register_offset(reg_name.lower())
            l.debug("Mapped register '%s' to offset %x", reg_name, reg_offset)
        except ValueError:
            reg_offset = var_data.offset + 0x100000
            l.debug("Could not map register '%s' from archinfo. Mapping to %x", reg_name, reg_offset)
        return reg_offset

    @staticmethod
    def _adjust_value_size(num_bits: int, v_in: BV) -> BV:
        """
        Ensure given bv is num_bits bits long by either zero extending or truncating.
        """
        if v_in.size() > num_bits:
            v_out = v_in[num_bits-1:0]
            l.debug('Truncating value %s (%d bits) to %s (%d bits)',
                    v_in,
                    v_in.size(),
                    v_out,
                    num_bits)
            return v_out
        elif v_in.size() < num_bits:
            v_out = v_in.zero_extend(num_bits-v_in.size())
            l.debug('Extending value %s (%d bits) to %s (%d bits)',
                    v_in,
                    v_in.size(),
                    v_out,
                    num_bits)
            return v_out
        else:
            return v_in

    def _set_value(self, var_data: VarnodeData, value: BV) -> None:
        """
        Store a value for a given varnode.

        This method stores to the appropriate register, or unique space,
        depending on the space indicated by the varnode.

        :param var_data: The varnode to store into.
        :param value:    The value to store.
        """
        space_name = var_data.space.getName()

        # FIXME: Consider moving into behavior.py
        value = self._adjust_value_size(var_data.size*8, value)
        assert var_data.size*8 == value.size()

        l.debug("Storing %s %x %s %d", space_name, var_data.offset, value, var_data.size)
        if space_name == "register":
            self.state.registers.store(
                self._map_register_name(var_data), value, size=var_data.size
            )

        elif space_name == "unique":
            self._pcode_tmps[var_data.offset] = value

        elif space_name in ("ram", "mem"):
            l.debug("Storing %s to offset %s", value, var_data.offset)
            self.state.memory.store(var_data.offset, value, endness=self.project.arch.memory_endness)

        else:
            raise NotImplementedError()

    def _get_value(self, var_data: VarnodeData) -> BV:
        """
        Get a value for a given varnode.

        This method loads from the appropriate const, register, unique, or RAM
        space, depending on the space indicated by the varnode.

        :param var_data: The varnode to load from.
        :return:         The value loaded.
        """
        space_name = var_data.space.getName()
        size = var_data.size
        l.debug("Loading %s - %x x %d", space_name, var_data.offset, size)
        if space_name == "const":
            return claripy.BVV(var_data.offset, size*8)
        elif space_name == "register":
            return self.state.registers.load(self._map_register_name(var_data), size=size)
        elif space_name == "unique":
            # FIXME: Support loading data of different sizes. For now, assume
            # size of values read are same as size written.
            assert self._pcode_tmps[var_data.offset].size() == size*8
            return self._pcode_tmps[var_data.offset]
        elif space_name in ("ram", "mem"):
            val = self.state.memory.load(var_data.offset, endness=self.project.arch.memory_endness, size=size)
            l.debug("Loaded %s from offset %s", val, var_data.offset)
            return val
        else:
            raise NotImplementedError()

    def _execute_unary(self) -> None:
        """
        Execute the unary behavior of the current op.
        """
        in1 = self._get_value(self._current_op.getInput(0))
        l.debug("in1 = %s", in1)
        out = self._current_behavior.evaluate_unary(
            self._current_op.getOutput().size, self._current_op.getInput(0).size, in1
        )
        l.debug("out unary = %s", out)
        self._set_value(self._current_op.getOutput(), out)

    def _execute_binary(self) -> None:
        """
        Execute the binary behavior of the current op.
        """
        in1 = self._get_value(self._current_op.getInput(0))
        in2 = self._get_value(self._current_op.getInput(1))
        l.debug("in1 = %s", in1)
        l.debug("in2 = %s", in2)
        out = self._current_behavior.evaluate_binary(
            self._current_op.getOutput().size, self._current_op.getInput(0).size, in1, in2
        )
        l.debug("out binary = %s", out)
        self._set_value(self._current_op.getOutput(), out)

    def _execute_load(self) -> None:
        """
        Execute a p-code load operation.
        """
        spc = pypcode.Address.getSpaceFromConst(self._current_op.getInput(0).getAddr())
        assert spc.getName() in ("ram", "mem")
        off = self._get_value(self._current_op.getInput(1))
        out = self._current_op.getOutput()
        res = self.state.memory.load(off, out.size, endness=self.project.arch.memory_endness)
        l.debug("Loaded %s from offset %s", res, off)
        self._set_value(out, res)

    def _execute_store(self) -> None:
        """
        Execute a p-code store operation.
        """
        spc = pypcode.Address.getSpaceFromConst(self._current_op.getInput(0).getAddr())
        assert spc.getName() in ("ram", "mem")
        off = self._get_value(self._current_op.getInput(1))
        data = self._get_value(self._current_op.getInput(2))
        l.debug("Storing %s at offset %s", data, off)
        self.state.memory.store(off, data, endness=self.project.arch.memory_endness)

    def _execute_branch(self) -> None:
        """
        Execute a p-code branch operation.
        """
        dest_addr = self._current_op.getInput(0).getAddr()
        if dest_addr.isConstant():
            raise NotImplementedError("P-code relative branch not supported yet")

        self.state.scratch.exit_handled = True
        expr = dest_addr.getOffset()
        self.successors.add_successor(
            self.state,
            expr,
            self.state.scratch.guard,
            "Ijk_Boring",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

    def _execute_cbranch(self) -> None:
        """
        Execute a p-code conditional branch operation.
        """
        cond = self._get_value(self._current_op.getInput(1))
        dest_addr = self._current_op.getInput(0).getAddr()
        if dest_addr.isConstant():
            raise NotImplementedError("P-code relative branch not supported yet")

        exit_state = self.state.copy()
        self.successors.add_successor(
            exit_state,
            dest_addr.getOffset(),
            cond != 0,
            "Ijk_Boring",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

        cont_state = self.state
        cont_condition = cond == 0
        cont_state.add_constraints(cont_condition)
        cont_state.scratch.guard = claripy.And(
            cont_state.scratch.guard, cont_condition
        )

    def _execute_ret(self) -> None:
        """
        Execute a p-code return operation.
        """
        self.state.scratch.exit_handled = True
        self.successors.add_successor(
            self.state,
            self.state.regs.ip,
            self.state.scratch.guard,
            "Ijk_Ret",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

    def _execute_branchind(self) -> None:
        """
        Execute a p-code indirect branch operation.
        """
        self.state.scratch.exit_handled = True
        expr = self._get_value(self._current_op.getInput(0))
        self.successors.add_successor(
            self.state,
            expr,
            self.state.scratch.guard,
            "Ijk_Boring",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

    def _execute_call(self) -> None:
        """
        Execute a p-code call operation.
        """
        self.state.scratch.exit_handled = True
        expr = self._current_op.getInput(0).getAddr().getOffset()
        self.successors.add_successor(
            self.state,
            expr,
            self.state.scratch.guard,
            "Ijk_Call",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

    def _execute_callind(self) -> None:
        """
        Execute a p-code indirect call operation.
        """
        self.state.scratch.exit_handled = True
        expr = self._get_value(self._current_op.getInput(0))
        self.successors.add_successor(
            self.state,
            expr,
            self.state.scratch.guard,
            "Ijk_Call",
            exit_stmt_idx=DEFAULT_STATEMENT,
            exit_ins_addr=self.state.scratch.ins_addr,
        )

    def _execute_callother(self) -> None:
        raise NotImplementedError("CALLOTHER emulation not currently supported")

    def _execute_multiequal(self) -> None:
        raise NotImplementedError("MULTIEQUAL appearing in unheritaged code?")

    def _execute_indirect(self) -> None:
        raise NotImplementedError("INDIRECT appearing in unheritaged code?")

    def _execute_segment_op(self) -> None:
        raise NotImplementedError("SEGMENTOP emulation not currently supported")

    def _execute_cpool_ref(self) -> None:
        raise NotImplementedError("Cannot currently emulate cpool operator")

    def _execute_new(self) -> None:
        raise NotImplementedError("Cannot currently emulate new operator")
